# -*- coding: utf-8 -*-
"""
The methods of the `NodeMixin` class should not access the user class special methods.

For instance, the user define a `MyNode` class as bellow:

```python
from anytree import NodeMixin


class MyNode(NodeMixin):
    def __init__(self, name, parent=None, children=None):
        super(MyNode, self).__init__()
        self.name = name
        self.parent = parent
        if children:
            self.children = children
```

In this class, the used can implement some special methods, like ``__eq__`` or ``__len__``,
which can have a specific meaning not related to the Tree structure.
A good exemple could be a `NodeMixin` subclass which also implements `collections.abc.Mapping`:


```python
import collections

from anytree import NodeMixin


class MyMapping(NodeMixin, collections.abc.Mapping):
    def __init__(self, name, parent=None, children=None):
        super(MyMapping, self).__init__()
        self.name = name
        self.parent = parent
        if children:
            self.children = children

    def __iter__(self):
        for child in self.children:
            yield child
            for item in child:
                yield item

    def __len__(self):
        return len(list(iter(self)))

    def __getitem__(self, name):
        for child in self:
            if child.name == name:
                return child
        raise KeyError(name)
```

In this example, the `NodeMixin` class shouldn't make any call to `__iter__`, `__len__` or `__getitem__`.

To avoid that, the `NodeMixin` class should respect the following rules:

- Do not compare nodes by value but by reference.
  In other words, the `NodeMixin` class should not use `==` or `!=` but compare nodes with the `id()` function.
  Comparison could be done with the `is` or `is not` operator. Avoid using `in` or `not in`.

- Do not use truth value testing to check if a node is "empty".
  The `NodeMixin` class should not `if node`, `if not node`, `while node` nor `while not node`.
  Instead, it must compare the node with `None` like this: `if node is not None`, `if node is None`,
  `while node is not None` or `while node is None`.

- Do not presume that nodes are hashable.
  The `NodeMixin` class should not store nodes in `set` or `dict`.
  Instead, it can store node IDs using the `id()` function.
"""
import functools
import unittest

from anytree import NodeMixin

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping


# List of method names to which we want to control access:
#
# - This list does not contain the access control methods (`__getattribute__`, `__getattr__`,
#   `__setattr__`, and `__delattr__`) which are used by the NodeMixin class anyway.
# - This list does not contain `__new__`, `__init__`, and `__del__` which are required for testing.
# - Some methods are not available in all Python version: `__bytes__`, `__unicode__` and `__dir__`.

SPECIAL_METHODS = [
    # rich comparison methods
    "__lt__",
    "__le__",
    "__eq__",
    "__ne__",
    "__gt__",
    "__ge__",
    # repr/str/bytes
    "__repr__",
    "__str__",
    # "__bytes__",
    # "__unicode__",
    "__format__",
    # hashing
    "__hash__",
    # attribute listing
    # "__dir__",
    # pickle protocol
    "__getnewargs_ex__",
    "__getnewargs__",
    "__getstate__",
    "__setstate__",
    "__reduce_ex__",
    # object size
    "__sizeof__",
    # truth value testing
    "__bool__",
    # callable objects
    "__call__",
    # container types
    "__len__",
    "__length_hint__",
    "__getitem__",
    "__setitem__",
    "__delitem__",
    "__missing__",
    "__iter__",
    "__reversed__",
    "__contains__",
    # numeric types
    "__add__",
    "__sub__",
    "__mul__",
    "__matmul__",
    "__truediv__",
    "__floordiv__",
    "__mod__",
    "__divmod__",
    "__pow__",
    "__lshift__",
    "__rshift__",
    "__and__",
    "__xor__",
    "__or__",
    "__radd__",
    "__rsub__",
    "__rmul__",
    "__rmatmul__",
    "__rtruediv__",
    "__rfloordiv__",
    "__rmod__",
    "__rdivmod__",
    "__rpow__",
    "__rlshift__",
    "__rrshift__",
    "__rand__",
    "__rxor__",
    "__ror__",
    "__iadd__",
    "__isub__",
    "__imul__",
    "__imatmul__",
    "__itruediv__",
    "__ifloordiv__",
    "__imod__",
    "__ipow__",
    "__ilshift__",
    "__irshift__",
    "__iand__",
    "__ixor__",
    "__ior__",
    "__neg__",
    "__pos__",
    "__abs__",
    "__invert__",
    "__complex__",
    "__int__",
    "__float__",
    "__index__",
    "__round__",
    "__trunc__",
    "__floor__",
    "__ceil__",
    # With Statement Context Managers
    "__enter__",
    "__exit__",
]
SPECIAL_METHODS += [attr for attr in ["__bytes__", "__unicode__", "__dir"] if hasattr(object, attr)]


def prevent_access(attr, *args, **kwargs):
    raise AssertionError("invalid call to " + attr)


class MyNode(NodeMixin):
    @staticmethod
    def __new__(cls, *args, **kwargs):
        for attr in SPECIAL_METHODS:
            setattr(cls, attr, functools.partial(prevent_access, attr))
        instance = super(NodeMixin, cls).__new__(cls)
        return instance

    def __init__(self, name, parent=None, children=None):
        super(MyNode, self).__init__()
        self.name = name
        self.parent = parent
        if children:
            self.children = children


class TestConsistency(unittest.TestCase):
    """Control the access to special methods"""

    def setUp(self):
        super(TestConsistency, self).setUp()
        self.root1 = MyNode("root1")
        self.child1 = MyNode("child1", parent=self.root1)
        self.child2a = MyNode("child2a", parent=self.child1)
        self.child2b = MyNode("child2b", parent=self.child1)
        self.other = MyNode("other")

    def test_parent__root1(self):
        _ = self.root1.parent

    def test_parent__setter__root1(self):
        self.root1.parent = self.other

    def test_children__root1(self):
        _ = self.root1.children

    def test_children__setter__root1(self):
        self.root1.children = [self.other]

    def test_path__root1(self):
        _ = self.root1.path

    def test_iter_path_reverse__root1(self):
        for _ in self.root1.iter_path_reverse():
            pass

    def test_ancestors__root1(self):
        _ = self.root1.ancestors

    def test_descendants__root1(self):
        _ = self.root1.descendants

    def test_root__root1(self):
        _ = self.root1.root

    def test_siblings__root1(self):
        _ = self.root1.siblings

    def test_leaves__root1(self):
        _ = self.root1.leaves

    def test_is_leaf__root1(self):
        _ = self.root1.is_leaf

    def test_is_root__root1(self):
        _ = self.root1.is_root

    def test_height__root1(self):
        _ = self.root1.height

    def test_depth__root1(self):
        _ = self.root1.depth

    def test_parent__child2b(self):
        _ = self.child2b.parent

    def test_parent__setter__child2b(self):
        self.child2b.parent = self.other

    def test_children__child2b(self):
        _ = self.child2b.children

    def test_children__setter__child2b(self):
        self.child2b.children = [self.other]

    def test_path__child2b(self):
        _ = self.child2b.path

    def test_iter_path_reverse__child2b(self):
        for _ in self.child2b.iter_path_reverse():
            pass

    def test_ancestors__child2b(self):
        _ = self.child2b.ancestors

    def test_descendants__child2b(self):
        _ = self.child2b.descendants

    def test_root__child2b(self):
        _ = self.child2b.root

    def test_siblings__child2b(self):
        _ = self.child2b.siblings

    def test_leaves__child2b(self):
        _ = self.child2b.leaves

    def test_is_leaf__child2b(self):
        _ = self.child2b.is_leaf

    def test_is_root__child2b(self):
        _ = self.child2b.is_root

    def test_height__child2b(self):
        _ = self.child2b.height

    def test_depth__child2b(self):
        _ = self.child2b.depth


class MyMapping(NodeMixin, Mapping):
    """
    This class is used to demonstrate a possible implementation
    which defines some special methods.
    """

    def __init__(self, name, parent=None, children=None):
        super(MyMapping, self).__init__()
        self.name = name
        self.parent = parent
        if children:
            self.children = children

    def __iter__(self):
        """Iterate over all children recursively."""
        for child in self.children:
            yield child
            for item in child:
                yield item

    def __len__(self):
        """Total number of children."""
        return len(list(iter(self)))

    def __getitem__(self, name):
        for child in self:
            if child.name == name:
                return child
        raise KeyError(name)


class TestMyMapping(unittest.TestCase):
    def setUp(self):
        super(TestMyMapping, self).setUp()
        self.root1 = MyMapping("root1")
        self.child1 = MyMapping("child1", parent=self.root1)
        self.child2a = MyMapping("child2a", parent=self.child1)
        self.child2b = MyMapping("child2b", parent=self.child1)

    def test_iter(self):
        expected_list = [self.child1, self.child2a, self.child2b]
        for actual, expected in zip(iter(self.root1), expected_list):
            self.assertIs(actual, expected)

    def test_len(self):
        self.assertEqual(len(self.root1), 3)

    def test_getitem(self):
        self.assertIs(self.root1["child1"], self.child1)
        self.assertIs(self.root1["child2a"], self.child2a)
        self.assertIs(self.root1["child2b"], self.child2b)
        with self.assertRaises(KeyError):
            _ = self.root1["missing"]



from anytree import (
    LevelGroupOrderIter,
    LevelOrderGroupIter,
    LevelOrderIter,
    Node,
    PostOrderIter,
    PreOrderIter,
    ZigZagGroupIter,
)

from .helper import eq_


def test_preorder():
    """PreOrderIter."""
    f = Node("f")
    b = Node("b", parent=f)
    a = Node("a", parent=b)
    d = Node("d", parent=b)
    c = Node("c", parent=d)
    e = Node("e", parent=d)
    g = Node("g", parent=f)
    i = Node("i", parent=g)
    h = Node("h", parent=i)

    eq_(list(PreOrderIter(f)), [f, b, a, d, c, e, g, i, h])
    eq_(list(PreOrderIter(f, maxlevel=0)), [])
    eq_(list(PreOrderIter(f, maxlevel=3)), [f, b, a, d, g, i])
    eq_(list(PreOrderIter(f, filter_=lambda n: n.name not in ("e", "g"))), [f, b, a, d, c, i, h])
    eq_(list(PreOrderIter(f, stop=lambda n: n.name == "d")), [f, b, a, g, i, h])

    it = PreOrderIter(f)
    eq_(next(it), f)
    eq_(next(it), b)
    eq_(list(it), [a, d, c, e, g, i, h])


def test_postorder():
    """PostOrderIter."""
    f = Node("f")
    b = Node("b", parent=f)
    a = Node("a", parent=b)
    d = Node("d", parent=b)
    c = Node("c", parent=d)
    e = Node("e", parent=d)
    g = Node("g", parent=f)
    i = Node("i", parent=g)
    h = Node("h", parent=i)

    eq_(list(PostOrderIter(f)), [a, c, e, d, b, h, i, g, f])
    eq_(list(PostOrderIter(f, maxlevel=0)), [])
    eq_(list(PostOrderIter(f, maxlevel=3)), [a, d, b, i, g, f])
    eq_(list(PostOrderIter(f, filter_=lambda n: n.name not in ("e", "g"))), [a, c, d, b, h, i, f])
    eq_(list(PostOrderIter(f, stop=lambda n: n.name == "d")), [a, b, h, i, g, f])

    it = PostOrderIter(f)
    eq_(next(it), a)
    eq_(next(it), c)
    eq_(list(it), [e, d, b, h, i, g, f])


def test_levelorder():
    """LevelOrderIter."""
    f = Node("f")
    b = Node("b", parent=f)
    a = Node("a", parent=b)
    d = Node("d", parent=b)
    c = Node("c", parent=d)
    e = Node("e", parent=d)
    g = Node("g", parent=f)
    i = Node("i", parent=g)
    h = Node("h", parent=i)

    eq_(list(LevelOrderIter(f)), [f, b, g, a, d, i, c, e, h])
    eq_(list(LevelOrderIter(f, maxlevel=0)), [])
    eq_(list(LevelOrderIter(f, maxlevel=3)), [f, b, g, a, d, i])
    eq_(list(LevelOrderIter(f, filter_=lambda n: n.name not in ("e", "g"))), [f, b, a, d, i, c, h])
    eq_(list(LevelOrderIter(f, stop=lambda n: n.name == "d")), [f, b, g, a, i, h])

    it = LevelOrderIter(f)
    eq_(next(it), f)
    eq_(next(it), b)
    eq_(list(it), [g, a, d, i, c, e, h])


def test_levelgrouporder():
    """LevelGroupOrderIter."""
    f = Node("f")
    b = Node("b", parent=f)
    a = Node("a", parent=b)
    d = Node("d", parent=b)
    c = Node("c", parent=d)
    e = Node("e", parent=d)
    g = Node("g", parent=f)
    i = Node("i", parent=g)
    h = Node("h", parent=i)

    eq_(list(LevelGroupOrderIter(f)), [(f,), (b, g), (a, d, i), (c, e, h)])
    eq_(list(LevelGroupOrderIter(f, maxlevel=0)), [])
    eq_(list(LevelGroupOrderIter(f, maxlevel=3)), [(f,), (b, g), (a, d, i)])
    eq_(list(LevelGroupOrderIter(f, filter_=lambda n: n.name not in ("e", "g"))), [(f,), (b,), (a, d, i), (c, h)])
    eq_(list(LevelGroupOrderIter(f, stop=lambda n: n.name == "d")), [(f,), (b, g), (a, i), (h,)])

    it = LevelGroupOrderIter(f)
    eq_(next(it), (f,))
    eq_(next(it), (b, g))
    eq_(list(it), [(a, d, i), (c, e, h)])


def test_levelordergroup():
    """LevelOrderGroupIter."""
    f = Node("f")
    b = Node("b", parent=f)
    a = Node("a", parent=b)
    d = Node("d", parent=b)
    c = Node("c", parent=d)
    e = Node("e", parent=d)
    g = Node("g", parent=f)
    i = Node("i", parent=g)
    h = Node("h", parent=i)

    eq_(list(LevelOrderGroupIter(f)), [(f,), (b, g), (a, d, i), (c, e, h)])
    eq_(list(LevelOrderGroupIter(f, maxlevel=0)), [])
    eq_(list(LevelOrderGroupIter(f, maxlevel=3)), [(f,), (b, g), (a, d, i)])
    eq_(list(LevelOrderGroupIter(f, filter_=lambda n: n.name not in ("e", "g"))), [(f,), (b,), (a, d, i), (c, h)])
    eq_(list(LevelOrderGroupIter(f, stop=lambda n: n.name == "d")), [(f,), (b, g), (a, i), (h,)])

    it = LevelOrderGroupIter(f)
    eq_(next(it), (f,))
    eq_(next(it), (b, g))
    eq_(list(it), [(a, d, i), (c, e, h)])


def test_zigzaggroup():
    """ZigZagGroupIter."""
    f = Node("f")
    b = Node("b", parent=f)
    a = Node("a", parent=b)
    d = Node("d", parent=b)
    c = Node("c", parent=d)
    e = Node("e", parent=d)
    g = Node("g", parent=f)
    i = Node("i", parent=g)
    h = Node("h", parent=i)

    eq_(list(ZigZagGroupIter(f)), [(f,), (g, b), (a, d, i), (h, e, c)])
    eq_(list(ZigZagGroupIter(f, maxlevel=0)), [])
    eq_(list(ZigZagGroupIter(f, maxlevel=3)), [(f,), (g, b), (a, d, i)])
    eq_(list(ZigZagGroupIter(f, filter_=lambda n: n.name not in ("e", "g"))), [(f,), (b,), (a, d, i), (h, c)])
    eq_(list(ZigZagGroupIter(f, stop=lambda n: n.name == "d")), [(f,), (g, b), (a, i), (h,)])

    it = ZigZagGroupIter(f)
    eq_(next(it), (f,))
    eq_(next(it), (g, b))
    eq_(list(it), [(a, d, i), (h, e, c)])



# -*- coding: utf-8 -*-
from enum import IntEnum

from pytest import raises

import anytree as at
from anytree import Node, Resolver

from .helper import assert_raises


def test_get():
    """Get."""
    top = at.Node("top", parent=None)
    sub0 = at.Node("sub0", parent=top)
    sub0sub0 = at.Node("sub0sub0", parent=sub0)
    sub0sub1 = at.Node("sub0sub1", parent=sub0)
    sub1 = at.Node("sub1", parent=top)
    r = at.Resolver("name")
    assert r.get(top, "sub0/sub0sub0") == sub0sub0
    assert r.get(sub1, "..") == top
    assert r.get(sub1, "../") == top
    assert r.get(sub1, "../.") == top
    assert r.get(sub1, "../sub0/sub0sub1") == sub0sub1
    assert r.get(sub1, ".") == sub1
    assert r.get(sub1, "") == sub1
    with assert_raises(at.ChildResolverError, "Node('/top') has no child sub2. Children are: 'sub0', 'sub1'."):
        r.get(top, "sub2")
    assert r.get(sub0sub0, "/top") == top
    assert r.get(sub0sub0, "/top/sub0") == sub0
    with assert_raises(at.RootResolverError, "Cannot go above root node Node('/top')"):
        r.get(top, "..")
    with assert_raises(at.ResolverError, "root node missing. root is '/top'."):
        r.get(sub0sub0, "/")
    with assert_raises(at.ResolverError, "unknown root node '/bar'. root is '/top'."):
        r.get(sub0sub0, "/bar")


def test_get_relaxed():
    """Get in relaxed mode."""
    top = at.Node("top", parent=None)
    sub0 = at.Node("sub0", parent=top)
    sub0sub0 = at.Node("sub0sub0", parent=sub0)
    sub0sub1 = at.Node("sub0sub1", parent=sub0)
    sub1 = at.Node("sub1", parent=top)
    r = at.Resolver("name", relax=True)
    assert r.get(top, "sub0/sub0sub0") == sub0sub0
    assert r.get(sub1, "..") == top
    assert r.get(sub1, "../") == top
    assert r.get(sub1, "../.") == top
    assert r.get(sub1, "../sub0/sub0sub1") == sub0sub1
    assert r.get(sub1, ".") == sub1
    assert r.get(sub1, "") == sub1
    assert r.get(top, "sub2") is None
    assert r.get(sub0sub0, "/top") == top
    assert r.get(sub0sub0, "/top/sub0") == sub0
    assert r.get(top, "..") is None
    assert r.get(sub0sub0, "/") is None
    assert r.get(sub0sub0, "/bar") is None


def test_glob():
    """Wildcard."""
    top = at.Node("top", parent=None)
    sub0 = at.Node("sub0", parent=top)
    sub0sub0 = at.Node("sub0", parent=sub0)
    sub0sub1 = at.Node("sub1", parent=sub0)
    sub0sub1sub0 = at.Node("sub0", parent=sub0sub1)
    at.Node("sub1", parent=sub0sub1)
    sub1 = at.Node("sub1", parent=top)
    sub1sub0 = at.Node("sub0", parent=sub1)
    r = at.Resolver()

    assert r.glob(top, "sub0/sub0") == [sub0sub0]
    assert r.glob(sub1, "..") == [top]
    assert r.glob(sub1, "../") == [top]
    assert r.glob(sub1, "../.") == [top]
    assert r.glob(sub1, "../././.") == [top]
    assert r.glob(sub1, ".././././sub0/..") == [top]
    assert r.glob(sub1, "../sub0/sub1") == [sub0sub1]
    assert r.glob(sub1, ".") == [sub1]
    assert r.glob(sub1, "./") == [sub1]
    assert r.glob(sub1, "") == [sub1]

    assert r.glob(sub1, "/top") == [top]
    assert r.glob(sub1, "/*") == [top]

    assert r.glob(top, "*/*/sub0") == [sub0sub1sub0]
    assert r.glob(top, "sub0/sub?") == [sub0sub0, sub0sub1]
    assert r.glob(sub1, ".././*") == [sub0, sub1]
    assert r.glob(top, "*/*") == [sub0sub0, sub0sub1, sub1sub0]
    assert r.glob(top, "*/sub0") == [sub0sub0, sub1sub0]
    with assert_raises(at.RootResolverError, "Cannot go above root node Node('/top')"):
        r.glob(top, "..")
    with assert_raises(at.RootResolverError, "Cannot go above root node Node('/top')"):
        r.glob(sub1, ".././..")
    with assert_raises(at.ChildResolverError, "Node('/top/sub1') has no child sub1. Children are: 'sub0'."):
        r.glob(top, "sub1/sub1")

    with assert_raises(at.ResolverError, "unknown root node '/z*'. root is '/top'."):
        r.glob(sub1, "/z*")

    # Recursive matching
    assert r.glob(top, "**/sub0") == [sub0, sub0sub0, sub0sub1sub0, sub1sub0]
    assert r.glob(top, "**/sub0/sub0") == [sub0sub0]
    assert r.glob(top, "**/**/sub0") == [sub0, sub0sub0, sub0sub1sub0, sub1sub0]
    assert r.glob(top, "sub0/**/sub0") == [sub0sub0, sub0sub1sub0]
    with assert_raises(at.ResolverError, "unknown root node '/sub0'. root is '/top'."):
        r.glob(top, "/sub0/**/sub0")


def test_glob_relaxed():
    """Wildcard relaxed."""
    top = at.Node("top", parent=None)
    sub0 = at.Node("sub0", parent=top)
    sub0sub0 = at.Node("sub0", parent=sub0)
    sub0sub1 = at.Node("sub1", parent=sub0)
    sub0sub1sub0 = at.Node("sub0", parent=sub0sub1)
    at.Node("sub1", parent=sub0sub1)
    sub1 = at.Node("sub1", parent=top)
    sub1sub0 = at.Node("sub0", parent=sub1)
    r = at.Resolver(relax=True)

    assert r.glob(top, "sub0/sub0") == [sub0sub0]
    assert r.glob(sub1, "..") == [top]
    assert r.glob(sub1, "../") == [top]
    assert r.glob(sub1, "../.") == [top]
    assert r.glob(sub1, "../././.") == [top]
    assert r.glob(sub1, ".././././sub0/..") == [top]
    assert r.glob(sub1, "../sub0/sub1") == [sub0sub1]
    assert r.glob(sub1, ".") == [sub1]
    assert r.glob(sub1, "./") == [sub1]
    assert r.glob(sub1, "") == [sub1]

    assert r.glob(sub1, "/top") == [top]
    assert r.glob(sub1, "/*") == [top]

    assert r.glob(top, "*/*/sub0") == [sub0sub1sub0]
    assert r.glob(top, "sub0/sub?") == [sub0sub0, sub0sub1]
    assert r.glob(sub1, ".././*") == [sub0, sub1]
    assert r.glob(top, "*/*") == [sub0sub0, sub0sub1, sub1sub0]
    assert r.glob(top, "*/sub0") == [sub0sub0, sub1sub0]
    assert r.glob(top, "..") == []
    assert r.glob(sub1, ".././..") == []
    assert r.glob(top, "sub1/sub1") == []

    assert r.glob(sub1, "/z*") == []

    # Recursive matching
    assert r.glob(top, "**/sub0") == [sub0, sub0sub0, sub0sub1sub0, sub1sub0]
    assert r.glob(top, "**/sub0/sub0") == [sub0sub0]
    assert r.glob(top, "**/**/sub0") == [sub0, sub0sub0, sub0sub1sub0, sub1sub0]
    assert r.glob(top, "sub0/**/sub0") == [sub0sub0, sub0sub1sub0]
    assert r.glob(top, "/sub0/**/sub0") == []


def test_glob_cache():
    """Wildcard Cache."""
    root = at.Node("root")
    sub0 = at.Node("sub0", parent=root)
    sub1 = at.Node("sub1", parent=root)
    r = at.Resolver()
    # strip down cache size
    at.resolver._MAXCACHE = 2
    at.Resolver._match_cache.clear()
    assert len(at.Resolver._match_cache) == 0
    assert r.glob(root, "sub0") == [sub0]
    assert len(at.Resolver._match_cache) == 1
    assert r.glob(root, "sub1") == [sub1]
    assert len(at.Resolver._match_cache) == 2
    assert r.glob(root, "sub*") == [sub0, sub1]
    assert len(at.Resolver._match_cache) == 1


def test_same_name():
    """Same Name."""
    root = at.Node("root")
    sub0 = at.Node("sub", parent=root)
    sub1 = at.Node("sub", parent=root)
    r = at.Resolver()
    assert r.get(root, "sub") == sub0
    assert r.glob(root, "sub") == [sub0, sub1]


def test_ignorecase():
    """Case insensitive resolver"""
    root = at.Node("root")
    sub0 = at.Node("sUB0", parent=root)
    sub1 = at.Node("sub1", parent=root)
    r = at.Resolver(ignorecase=True)

    assert r.get(root, "SUB0") == sub0
    assert r.get(root, "sub0") == sub0
    assert r.get(root, "sUB0") == sub0
    assert r.glob(root, "SU*1") == [sub1]
    assert r.glob(root, "/*/SU*1") == [sub1]
    assert r.glob(sub0, "../*1") == [sub1]


def test_enum():
    class Animals(IntEnum):
        Mammal = 1
        Cat = 2
        Dog = 3

    root = Node("ANIMAL")
    mammal = Node(Animals.Mammal, parent=root)
    cat = Node(Animals.Cat, parent=mammal)
    dog = Node(Animals.Dog, parent=mammal)

    r = Resolver()
    assert r.glob(root, "/ANIMAL/*") == [mammal]
    assert r.glob(root, "/ANIMAL/*/*") == [cat, dog]


def test_glob_consistency():
    """Ensure Consistency"""
    node = at.Node("root")
    resolver = at.Resolver()
    assert resolver.glob(node, "/root") == [node]
    assert resolver.get(node, "/root") == node

    assert resolver.glob(node, ".") == [node]
    assert resolver.get(node, ".") == node

    assert resolver.glob(node, "") == [node]
    assert resolver.get(node, "") == node

    with raises(at.ResolverError):
        assert resolver.glob(node, "/")
    with raises(at.ResolverError):
        assert resolver.get(node, "/")

    with raises(at.RootResolverError):
        assert resolver.glob(node, "..")
    with raises(at.RootResolverError):
        assert resolver.get(node, "..")

    with raises(at.ResolverError):
        assert resolver.glob(node, "root")
    with raises(at.ResolverError):
        assert resolver.get(node, "root")

    with raises(at.ResolverError):
        assert resolver.glob(node, "bla")
    with raises(at.ResolverError):
        assert resolver.get(node, "bla")

    with raises(at.ResolverError):
        assert resolver.glob(node, "/bla")
    with raises(at.ResolverError):
        assert resolver.get(node, "/bla")


def test_glob_consistency_relax():
    """Ensure Consistency on relaxed."""
    node = at.Node("root")
    resolver = at.Resolver(relax=True)
    assert resolver.glob(node, "/root") == [node]
    assert resolver.get(node, "/root") == node

    assert resolver.glob(node, ".") == [node]
    assert resolver.get(node, ".") == node

    assert resolver.glob(node, "") == [node]
    assert resolver.get(node, "") == node

    assert resolver.glob(node, "/") == []
    assert resolver.get(node, "/") is None

    assert resolver.glob(node, "..") == []
    assert resolver.get(node, "..") is None

    assert resolver.glob(node, "root") == []
    assert resolver.get(node, "root") is None

    assert resolver.glob(node, "bla") == []
    assert resolver.get(node, "bla") is None

    assert resolver.glob(node, "/bla") == []
    assert resolver.get(node, "/bla") is None
